import { Backend, DatabaseConnection, DatabaseTransaction, Schema, RecordStoreRequest, RecordGetRequest, RecordGetResponse, RecordStoreResponse } from "./backend-interface";
import { ISortedMapF } from "./tree/b+tree";
import { Key, Value } from "./util/types";
import BridgeIDBKeyRange from "./BridgeIDBKeyRange";
declare enum TransactionLevel {
    Disconnected = 0,
    Connected = 1,
    Read = 2,
    Write = 3,
    VersionChange = 4
}
interface ObjectStore {
    originalName: string;
    modifiedName: string | undefined;
    originalData: ISortedMapF<Key, ObjectStoreRecord>;
    modifiedData: ISortedMapF<Key, ObjectStoreRecord> | undefined;
    deleted: boolean;
    originalKeyGenerator: number;
    modifiedKeyGenerator: number | undefined;
}
interface Index {
    originalName: string;
    modifiedName: string | undefined;
    originalData: ISortedMapF<Key, IndexRecord>;
    modifiedData: ISortedMapF<Key, IndexRecord> | undefined;
    deleted: boolean;
}
interface Database {
    committedObjectStores: {
        [name: string]: ObjectStore;
    };
    modifiedObjectStores: {
        [name: string]: ObjectStore;
    };
    committedIndexes: {
        [name: string]: Index;
    };
    modifiedIndexes: {
        [name: string]: Index;
    };
    committedSchema: Schema;
    /**
     * Was the transaction deleted during the running transaction?
     */
    deleted: boolean;
    txLevel: TransactionLevel;
    connectionCookie: string | undefined;
}
interface Connection {
    dbName: string;
    modifiedSchema: Schema | undefined;
    /**
     * Has the underlying database been deleted?
     */
    deleted: boolean;
    /**
     * Map from the effective name of an object store during
     * the transaction to the real name.
     */
    objectStoreMap: {
        [currentName: string]: ObjectStore;
    };
    indexMap: {
        [currentName: string]: Index;
    };
}
interface IndexRecord {
    indexKey: Key;
    primaryKeys: Key[];
}
interface ObjectStoreRecord {
    primaryKey: Key;
    value: Value;
}
declare class AsyncCondition {
    _waitPromise: Promise<void>;
    _resolveWaitPromise: () => void;
    constructor();
    wait(): Promise<void>;
    trigger(): void;
}
/**
 * Primitive in-memory backend.
 */
export declare class MemoryBackend implements Backend {
    databases: {
        [name: string]: Database;
    };
    connectionIdCounter: number;
    transactionIdCounter: number;
    /**
     * Connections by connection cookie.
     */
    connections: {
        [name: string]: Connection;
    };
    /**
     * Connections by transaction (!!) cookie.  In this implementation,
     * at most one transaction can run at the same time per connection.
     */
    connectionsByTransaction: {
        [tx: string]: Connection;
    };
    /**
     * Condition that is triggered whenever a client disconnects.
     */
    disconnectCond: AsyncCondition;
    /**
     * Conditation that is triggered whenever a transaction finishes.
     */
    transactionDoneCond: AsyncCondition;
    enableTracing: boolean;
    getDatabases(): Promise<{
        name: string;
        version: number;
    }[]>;
    deleteDatabase(tx: DatabaseTransaction, name: string): Promise<void>;
    connectDatabase(name: string): Promise<DatabaseConnection>;
    beginTransaction(conn: DatabaseConnection, objectStores: string[], mode: import("./util/types").TransactionMode): Promise<DatabaseTransaction>;
    enterVersionChange(conn: DatabaseConnection, newVersion: number): Promise<DatabaseTransaction>;
    close(conn: DatabaseConnection): Promise<void>;
    getSchema(dbConn: DatabaseConnection): Schema;
    renameIndex(btx: DatabaseTransaction, oldName: string, newName: string): void;
    deleteIndex(btx: DatabaseTransaction, indexName: string): void;
    deleteObjectStore(btx: DatabaseTransaction, name: string): void;
    renameObjectStore(btx: DatabaseTransaction, oldName: string, newName: string): void;
    createObjectStore(btx: DatabaseTransaction, name: string, keyPath: string | string[] | null, autoIncrement: boolean): void;
    createIndex(btx: DatabaseTransaction, indexName: string, objectStoreName: string, keyPath: import("./util/types").KeyPath, multiEntry: boolean, unique: boolean): void;
    deleteRecord(btx: DatabaseTransaction, objectStoreName: string, range: BridgeIDBKeyRange): Promise<void>;
    private deleteFromIndex;
    getRecords(btx: DatabaseTransaction, req: RecordGetRequest): Promise<RecordGetResponse>;
    storeRecord(btx: DatabaseTransaction, storeReq: RecordStoreRequest): Promise<RecordStoreResponse>;
    private insertIntoIndex;
    rollback(btx: DatabaseTransaction): Promise<void>;
    commit(btx: DatabaseTransaction): Promise<void>;
}
export default MemoryBackend;
